/*
 * On Demand WebSrc -- SourceCode to HTML converter
 * All rights reserved. Copyright (C) 1996 by narita
 */

#include <stdio.h>
#include <unistd.h>
#include <dbm.h>

/**********************************************************************/

#define DBFILE_WORD	"db"
#define KEY_WORD	"key"
#define PARTKEY_WORD	"partkey"
#define FILE_WORD	"file"
#define PARTIAL_WORD	"partial"

#define YES_WORD	"yes"
#define NO_WORD		"no"

/**********************************************************************/

#define VERSION		"1.05beta"

#define NULL		0

#define FALSE		0
#define TRUE		1

#define BUF_SIZE        256

#define DB_MERGE	0
#define DB_LOAD		1
#define DB_FETCH	2
#define DB_SELECT	3
#define DB_HTML		4
#define DB_SHOW		5

char *CGI_NAME         = "websrc";
char *DEFAULT_DATABASE = "database";
char *SEARCHER_REF     = "index.html";
char *SEARCHER_TITLE   = "No Title";
char *DEFAULT_ORIGIN   = "";

/**********************************************************************/

char *indexLookup( index )
  int index;
{
  datum keySlot;
  static datum contentSlot;
  char tmpKey[ BUF_SIZE ];

  sprintf( tmpKey, "#%d", index );
  keySlot.dptr = tmpKey;
  keySlot.dsize = strlen ( tmpKey );

  contentSlot = fetch( keySlot );

  return (char *)contentSlot.dptr;
}

typedef struct DATUM_CONTENT_NUMBER_TAG {
  int index;
  int line;
} datum_content_number_t;

typedef union DATUM_CONTENT_TAG {
  datum_content_number_t number;
  char dir[ BUF_SIZE ];
} datum_content_t;

typedef struct CONTENT_TAG {
  int index;
  int line;
  char *dir;
} content_t;

typedef struct CONTENT_LIST_TAG {
  content_t *content;
  struct CONTENT_LIST_TAG *next;
  char *key;
} content_list_t;

content_t *contentAlloc( index, line )
  int index;
  int line;
{
  content_t *content;

  if( NULL == (content = (content_t *)malloc( sizeof( content_t ) )) )
    return NULL;

  content->index = index;
  content->line  = line;
  content->dir   = NULL;

  return content;
}

content_t *contentExpand( content )
  content_t *content;
{
  char *dir;

  if( NULL != content->dir )
    return NULL;

  if( NULL == (dir = indexLookup( content->index )) )
    return NULL;

  if( NULL == (content->dir = (char *)malloc( 1 + strlen( dir ) )) )
    return NULL;

  strcpy( content->dir, dir );

  return content;
}

contentFree( content )
  content_t *content;
{
  if( content->dir )
    free( content->dir );
  free( content );
}

content_list_t *contentListAlloc( content, key )
  content_t *content;
  char *key;
{
  content_list_t *contentList;

  if( NULL == (contentList = (content_list_t *)malloc( sizeof( content_list_t ) )) )
    return NULL;

  if( key ){
    if( NULL == (contentList->key = (char *)malloc( 1 + strlen( key ) )) ){
      free( contentList );
      return NULL;
    }
    strcpy( contentList->key, key );
  } else {
    contentList->key = NULL;
  }
  contentList->content = content;
  contentList->next    = NULL;

  return contentList;
}

content_list_t *contentListAdd( contentList, content, key )
  content_list_t *contentList;
  content_t *content;
  char *key;
{
  content_list_t *newContentList;

  if( NULL == (newContentList = (content_list_t *)malloc( sizeof( content_list_t ) )) )
    return NULL;

  if( key ){
    if( NULL == (newContentList->key = (char *)malloc( 1 + strlen( key ) )) ){
      free( newContentList );
      return NULL;
    }
    strcpy( newContentList->key, key );
  } else {
    newContentList->key = NULL;
  }
  newContentList->content = content;
  newContentList->next    = contentList;

  return newContentList;
}

content_list_t *contentListExpand( contentList )
  content_list_t *contentList;
{
  content_t *content;
  content_list_t *res;

  res = contentList;

  for( ; contentList ; contentList = contentList->next )
    contentExpand( contentList->content );

  return res;
}

contentListFree( contentList )
  content_list_t *contentList;
{
  if( contentList->key )
    free( contentList->key );
  free( contentList );
}

contentListFreeAll( contentList )
  content_list_t *contentList;
{
  content_t *content;
  content_list_t *nextContentList;

  while( contentList ){
    nextContentList = contentList->next;
    content = contentList->content;
    contentFree( content );
    contentListFree( contentList );
    contentList = nextContentList;
  }
}

contentListPrint( contentList, key )
  content_list_t *contentList;
  char *key;
{
  content_t *content;

  if( NULL == contentList ){
    printf( "No entry\n" );
    return;
  }

  while( contentList ){
    content = contentList->content;
    printf( "%s %d %s\n",
	   ( contentList->key ? contentList->key : key ),
	   content->line,
	   content->dir );
    contentList = contentList->next;
  }
}

contentListHTML( contentList, key, dbfile )
  content_list_t *contentList;
  char *key;
  char *dbfile;
{
  content_t *content;

  if( NULL == contentList ){
    printf( "No entry\n" );
    return;
  }

  while( contentList ){
    content = contentList->content;
    printf( "<H3><LI>%s%s <A HREF=%s?-k%s+-l%d+-d%s+%s>%s (line %d)</A></H3>\n",
	   ( contentList->key ? contentList->key : "" ),
	   ( contentList->key ? "<BR>" : "" ),
	   CGI_NAME,
	   ( contentList->key ? contentList->key : key ),
	   content->line,
	   dbfile,
	   content->dir,
	   content->dir,
	   content->line );
    contentList = contentList->next;
  }
}

contentListCount( contentList )
  content_list_t *contentList;
{
  content_t *content;
  int count = 0;

  while( contentList ){
    count++;
    contentList = contentList->next;
  }

  return count;
}

/**********************************************************************/

#define FATAL_ERROR		0
#define NO_DATABASE		1
#define NO_KEYWORD		2
#define NO_FILE			3
#define NO_ENTRY		4

htmlError( err, str )
  int err;
  char *str;
{
  if( NULL == str )
    str = "";

  printf( "Content-type: text/html\n\n" );
  printf( "<HTML><HEAD><TITLE>Error</TITLE></HEAD>\n" );
  printf( "<BODY><HR>\n" );

  switch( err ){
  case NO_DATABASE:
    printf( "<H2>Search error occured:</H2><P>\n" );
    printf( "<H3><DL><DT><DD>No such database: ``%s''.</DL></H3>\n", str );
    break;
  case NO_KEYWORD:
    printf( "<H2>Search error occured:</H2><P>\n" );
    printf( "<H3><DL><DT><DD>No keyword.</DL></H3>\n" );
    break;
  case NO_FILE:
    printf( "<H2>Search error occured:</H2><P>\n" );
    printf( "<H3><DL><DT><DD>No such file: ``%s''.</DL></H3>\n", str );
    break;
  case NO_ENTRY:
    printf( "<H2>Search error occured:</H2><P>\n" );
    printf( "<H3><DL><DT><DD>No entry: ``%s''.</DL></H3>\n", str );
    break;
  default:
    printf( "<H2>Fatal error occured:</H2><P>\n" );
    printf( "<H3><DL><DT><DD>Process cannot be proceeded.</DL></H3>\n" );
    break;
  }

  printf( "<H3><A HREF=%s>Return to: ``%s''</A></H3>\n", SEARCHER_REF, SEARCHER_TITLE );

  printf( "<HR></BODY></HTML>\n" );
}

/**********************************************************************/

legalFile( file )
  char *file;
{
  char buf[ BUF_SIZE ];
  datum keySlot, contentSlot;

  if( NULL == file )
    return FALSE;

  strcpy( buf, "#" );
  strcat( buf, file );

  keySlot.dptr = buf;
  keySlot.dsize = strlen( buf );

  contentSlot = fetch( keySlot );
  if( NULL == contentSlot.dptr )
    return FALSE;

  return TRUE;
}

existDatabase( dbfile )
  char *dbfile;
{
  char buf[ BUF_SIZE ];

  if( NULL == dbfile )
    return FALSE;

  strcpy( buf, dbfile );
  strcat( buf, ".dir" );
  if( 0 > access( buf, R_OK ) )
    return FALSE;

  strcpy( buf, dbfile );
  strcat( buf, ".pag" );
  if( 0 > access( buf, R_OK ) )
    return FALSE;

  return TRUE;
}

touchDatabase( dbfile )
  char *dbfile;
{
  char buf[ BUF_SIZE ];

  if( NULL == dbfile )
    return FALSE;

  strcpy( buf, dbfile );
  strcat( buf, ".dir" );
  fclose( fopen( buf, "w" ) );
  strcpy( buf, dbfile );
  strcat( buf, ".pag" );
  fclose( fopen( buf, "w" ) );

  return TRUE;
}

dbMerge( file )
  char *file;
{
  FILE *fp;
  char path[ BUF_SIZE ], key[ BUF_SIZE ], tmpKey[ BUF_SIZE ];
  char *ptr, *line, *dir;
  datum_content_t datumContent;
  datum keySlot, contentSlot, oldContentSlot;
  int len, gen;
  int index, maxIndex;
  int flagThereWasIndex, flagAddDir;
  char intChar[ 4 ];

  strcpy( path, DEFAULT_ORIGIN );
  strcat( path, file );
  if( NULL == (fp = fopen( path, "r" )) )
    fprintf( stderr, "Cannot open: ``%s''\n", path ), exit( -1 );

  strcpy( tmpKey, "#INDEX" );
  keySlot.dptr = tmpKey;
  keySlot.dsize = strlen ( tmpKey );

  oldContentSlot = fetch( keySlot );
  if( NULL == oldContentSlot.dptr ){
    flagThereWasIndex = FALSE;
    maxIndex = 0;
  } else {
    flagThereWasIndex = TRUE;
    memcpy( intChar, oldContentSlot.dptr, sizeof( int ) );
    maxIndex = *(int *)intChar;
  }

#ifdef DEBUG
  printf( "Initial index: %d\n", maxIndex );
#endif DEBUG

  flagAddDir = FALSE;
  for( ; ; ){
    if( NULL == fgets( key, BUF_SIZE, fp ) )
      break;

    ptr = key;
    if( ptr = (char *)strchr( ptr, ' ' ) )
      *ptr = NULL, line = ++ptr;
    else if( '/' == *key ){
      if( ptr = (char *)strchr( key, 0x0a ) )
	*ptr = NULL;

      flagAddDir = TRUE;
      dir = key;
    } else
      fprintf( stderr, "Format error\n" ), exit( -1 );

    if( FALSE == flagAddDir ){
      if( ptr = (char *)strchr( ptr, ' ' ) )
	*ptr = NULL, dir = ++ptr;
      else
	fprintf( stderr, "Format error\n" ), exit( -1 );
      if( ptr = (char *)strchr( ptr, 0x0a ) )
	*ptr = NULL;
    }

    strcpy( tmpKey, "#" );
    strcat( tmpKey, dir );
    keySlot.dptr = tmpKey;
    keySlot.dsize = strlen( tmpKey );

    oldContentSlot = fetch( keySlot );
    if( NULL == oldContentSlot.dptr ){
      index = ++maxIndex;

      datumContent.number.index = index;
      contentSlot.dptr = (char *)&datumContent;
      contentSlot.dsize = sizeof( int );
      store( keySlot, contentSlot );

      sprintf( tmpKey, "#%d", index );
      keySlot.dptr = tmpKey;
      keySlot.dsize = strlen( tmpKey );

      strcpy( datumContent.dir, dir );
      contentSlot.dptr = (char *)&datumContent;
      contentSlot.dsize = 1 + strlen( datumContent.dir );
      store( keySlot, contentSlot );

      if( flagAddDir ){
	flagAddDir = FALSE;
	continue;
      }
    } else if( flagAddDir ){
      flagAddDir = FALSE;
      continue;
    } else {
      memcpy( intChar, oldContentSlot.dptr, sizeof( int ) );
      index = *(int *)intChar;
    }

    datumContent.number.index = index;
    datumContent.number.line = atoi( line );
    contentSlot.dptr = (char *)&datumContent;
    contentSlot.dsize = sizeof( datum_content_number_t );

    len = strlen( key );
    strcpy( tmpKey, key );
    keySlot.dptr = tmpKey;
    keySlot.dsize = len+1;

#ifdef DEBUG
    printf( "KEY:%s, INDEX:%d, FILE:%s, LINE:%d\n",
	   keySlot.dptr,
	   datumContent.number.index,
	   dir,
	   datumContent.number.line );
#endif DEBUG

    for( gen = 0 ; ; gen++ ){
      tmpKey[ len ] = ' ' + gen;

      oldContentSlot = fetch( keySlot );
      if( NULL == oldContentSlot.dptr ){
	store( keySlot, contentSlot );
	break;
      }
    }
  }

  strcpy( tmpKey, "#INDEX" );
  keySlot.dptr = tmpKey;
  keySlot.dsize = strlen ( tmpKey );

  if( flagThereWasIndex )
    delete( keySlot );

#ifdef DEBUG
  printf( "Current index is %d\n", maxIndex );
#endif DEBUG

  datumContent.number.index = maxIndex;
  contentSlot.dptr = (char *)&datumContent;
  contentSlot.dsize = sizeof( int );
  store( keySlot, contentSlot );
}

content_list_t *dbFetch( key )
  char *key;
{
  char intChar[ 4 ];
  char tmpKey[ BUF_SIZE ];
  datum keySlot, contentSlot;
  int len, gen;
  content_t *content;
  content_list_t *contentList, *newContentList;
  int flagFirst = TRUE;
  int index, line;

  for( gen = 0 ; ; gen++ ){
    len = strlen( key );
    strcpy( tmpKey, key );
    tmpKey[ len ] = ' ' + gen;

    keySlot.dptr = tmpKey;
    keySlot.dsize = len+1;

    contentSlot = fetch( keySlot );
    if( NULL == contentSlot.dptr ){
      if( flagFirst )
	return NULL;
      else
	return contentListExpand( contentList );
    }

    memcpy( intChar, &((datum_content_t *)contentSlot.dptr)->number.index,
	   sizeof( int ) );
    index = *(int *)intChar;
    memcpy( intChar, &((datum_content_t *)contentSlot.dptr)->number.line,
	   sizeof( int ) );
    line = *(int *)intChar;

    if( NULL == (content = contentAlloc( index, line )) )
      return NULL;

    if( flagFirst ){
      flagFirst = FALSE;
      if( NULL == (contentList = contentListAlloc( content, NULL )) ){
	contentFree( content );
	return NULL;
      }
    } else {
      if( NULL == (newContentList = contentListAdd( contentList,
						   content,
						   NULL )) ){
	contentFree( content );
	return contentListExpand( contentList );
      } else {
	contentList = newContentList;
      }
    }
  }
}

content_list_t *dbSearch( key )
  char *key;
{
  char intChar[ 4 ];
  datum keySlot, contentSlot;
  int len, gen;
  content_t *content;
  content_list_t *contentList = NULL, *newContentList;
  char tmpCh, *ptr;
  int flagFirst = TRUE;
  char completion[ BUF_SIZE ];
  int line, index;
  int count = 0;

  for( keySlot = firstkey() ; keySlot.dptr ; keySlot = nextkey( keySlot ) ){
    if( '#' == keySlot.dptr[ 0 ] || 1 > keySlot.dsize )
      continue;

    tmpCh = keySlot.dptr[ keySlot.dsize - 1 ];
    keySlot.dptr[ keySlot.dsize - 1 ] = NULL;

    if( !strstr( keySlot.dptr, key ) ){
      keySlot.dptr[ keySlot.dsize - 1 ] = tmpCh;
    } else {
      if( strlen( keySlot.dptr ) >= BUF_SIZE ){
	htmlError( FATAL_ERROR, NULL );
	exit( 0 );
      }

      strcpy( completion, keySlot.dptr );
      keySlot.dptr[ keySlot.dsize - 1 ] = tmpCh;

#ifdef DEBUG
      printf( "Found:%s\n", completion );
/*      continue;*/
#endif DEBUG

      contentSlot = fetch( keySlot );
      if( NULL == contentSlot.dptr ){
	if( flagFirst )
	  return NULL;
	else
	  return contentListExpand( contentList );
      }

      memcpy( intChar, &((datum_content_t *)contentSlot.dptr)->number.index,
	     sizeof( int ) );
      index = *(int *)intChar;
      memcpy( intChar, &((datum_content_t *)contentSlot.dptr)->number.line,
	     sizeof( int ) );
      line = *(int *)intChar;

/*** You should not involve below.
      printf( "KEY:%s, LINE:%d, INDEX:%d, DIR:%s\n",
	     completion,
	     line,
	     index,
	     indexLookup( index ) );
***/

      if( NULL == (content = contentAlloc( index, line )) )
	return NULL;

      if( flagFirst ){
	flagFirst = FALSE;
	if( NULL == (contentList = contentListAlloc( content,
						    completion )) ){
	  contentFree( content );
	  return NULL;
	}
      } else {
	if( NULL == (newContentList = contentListAdd( contentList,
						     content,
						     completion )) ){
	  contentFree( content );
	  return contentListExpand( contentList );
	} else {
	  contentList = newContentList;
	}
      }
    }
  }

  return contentListExpand( contentList );
}

/**********************************************************************/

hex2dec( c )
  char c;
{
  if( c >= '0' && c <= '9' )
    return c - '0';
  else if( c >= 'A' && c <= 'F' )
    return c - 'A' + 10;
  else if( c >= 'a' && c <= 'f' )
    return c - 'a' + 10;
  else
    return 0;
}

percentDecode( str )
  char *str;
{
  int len, i, j;
  char *ptr;

  if( !str )
    return -1;

  ptr = str;

  for( len = 0 ; *str = *ptr ; len++ ){
    if( '%' != *ptr ){
      *str++ = *ptr++;
    } else {
      if( *(ptr + 1) && *(ptr + 2) ){
	*str++ = ( hex2dec( *(ptr + 1) ) << 4 ) | hex2dec( *(ptr + 2) );
	ptr += 3;
      } else {
	htmlError();
	exit( 0 );
      }
    }
  }

  return len;
}

typedef struct QUERY_REPLY_TAG {
  char file[ BUF_SIZE ];
  char dbfile[ BUF_SIZE ];
  char key[ BUF_SIZE ];
  content_list_t *contentList;
} query_reply_t;

query_reply_t *readQuery()
{
  static query_reply_t queryReply;
  int i;
  char query[ BUF_SIZE ], val[ BUF_SIZE ], var[ BUF_SIZE ];
  char *ptr, *valPtr, *varPtr;
  int flagValue, flagPartial = FALSE;

  strcpy( queryReply.dbfile, DEFAULT_DATABASE );
  queryReply.key[ 0 ] = NULL;
  queryReply.file[ 0 ] = NULL;

  ptr = (char *)getenv( "QUERY_STRING" );
  strcpy( query, ptr );
  percentDecode( query );
  strcat( query, "&" );

#ifdef DEBUG
  printf( "Content-type: text/plain\n\n" );
  printf( "query: [%s]\n", query );
#endif DEBUG

  ptr = query;
  varPtr = var;
  valPtr = val;
  flagValue = FALSE;
  while( *ptr ){
    if( *ptr == '=' ){
      ptr++;
      *varPtr = NULL;
      flagValue = TRUE;
      varPtr = var;
    } else if( *ptr == '&' ){
      ptr++;
      *valPtr = NULL;
      flagValue = FALSE;
      valPtr = val;

      if( strlen( val ) < BUF_SIZE ){
	if( !strcmp( var, DBFILE_WORD ) )
	  strcpy( queryReply.dbfile, val );
	else if( !strcmp( var, KEY_WORD ) )
	  strcpy( queryReply.key, val );
	else if( !strcmp( var, FILE_WORD ) )
	  strcpy( queryReply.file, val );
	else if( !strcmp( var, PARTKEY_WORD ) ){
	  flagPartial = TRUE;
	  strcpy( queryReply.key, val );
	} else if( !strcmp( var, PARTIAL_WORD ) ){
	  if( !strcmp( val, YES_WORD ) )
	    flagPartial = TRUE;
	}
      }
    } else {
      if( flagValue )
	*valPtr++ = *ptr++;
      else
	*varPtr++ = *ptr++;
    }
  }

#ifdef DEBUG
  printf( "key: [%s]\n", queryReply.key );
  printf( "database: [%s]\n", queryReply.dbfile );
#endif DEBUG

  if( FALSE == existDatabase( queryReply.dbfile ) ){
    htmlError( NO_DATABASE, queryReply.dbfile );
    exit( 0 );
  }

  if( 0 > dbminit( queryReply.dbfile ) ){
    htmlError( NO_DATABASE, queryReply.dbfile );
    exit( 0 );
  }

  if( queryReply.key[ 0 ] ){
    if( flagPartial )
      queryReply.contentList = dbSearch( queryReply.key, queryReply.dbfile );
    else
      queryReply.contentList = dbFetch( queryReply.key, queryReply.dbfile );
  } else {
    queryReply.contentList = NULL;
  }

  return &queryReply;
}

/**********************************************************************/

#define htmlChar( ch, pos, line, count, flagLineNumber, flagAlwaysReload )\
{									\
  switch( ch ){								\
  case '<':								\
    pos++;								\
    printf( "&lt;" );							\
    break;								\
  case '>':								\
    pos++;								\
    printf( "&gt;" );							\
    break;								\
  case '"':								\
    pos++;								\
    printf( "&quot;" );							\
    break;								\
  case '&':								\
    pos++;								\
    printf( "&amp;" );							\
    break;								\
  case 0x09:								\
    {									\
      int i = 8 - pos % 8;						\
      pos += i;								\
      while( i-- > 0 )							\
        putchar( ' ' );							\
    }									\
    break;								\
  case 0x0a:								\
    printf( "\n" );							\
    pos = 0;								\
    (count)++;								\
    if( FALSE == flagAlwaysReload || line == count )			\
      printf( "<A NAME=%d>", (count) );					\
    if( flagLineNumber )						\
      printf( "%4d:", (count) );					\
    printf( " " );							\
    if( FALSE == flagAlwaysReload || line == count )			\
      printf( "</A>" );							\
    break;								\
  default:								\
    pos++;								\
    putchar( ch );							\
  }									\
}

#define isAlphabeticElement(c)	( ((c) >= 'a' && (c) <= 'z') || ((c) >= 'A' && (c) <= 'Z') || ((c) == '_') )

#define isWordElement(c)	( ((c) >= '0' && (c) <= '9') || ((c) >= 'a' && (c) <= 'z') || ((c) >= 'A' && (c) <= 'Z') || ((c) == '_') )

load( file, path, dbfile, line, flagLineNumber, flagAlwaysReload )
  char *file;
  char *path;
  char *dbfile;
  int line;
  int flagLineNumber;
  int flagAlwaysReload;
{
  FILE *fp;
  int ch;
  char wptr, word[ BUF_SIZE ];
  int flagWordHead;
  content_list_t *contentList;
  int count;
  int fileLen;
  int pos;

  fileLen = strlen( file );

  if( NULL == (fp = fopen( path, "r" )) ){
    printf( "Cannot open: ``%s''\n", path );
    printf( "</PRE><HR></BODY></HTML>\n" );
    exit( 0 );
  }

  line = line;
  count = 1;

  if( FALSE == flagAlwaysReload || line == count )
    printf( "<A NAME=%d>", (count) );
  if( flagLineNumber )
    printf( "%4d:", (count) );
  printf( " " );
  if( FALSE == flagAlwaysReload || line == count )
    printf( "</A>" );

  wptr = 0;
  pos = 0;
  flagWordHead = TRUE;
  for( ; ; ){
    if( -1 == (ch = fgetc( fp )) )
      break;
    if( flagWordHead ){
      if( isAlphabeticElement( ch ) ){
	flagWordHead = FALSE;
	wptr = 0;
	word[ wptr++ ] = ch;
      } else {
	htmlChar( ch, pos, line, count, flagLineNumber, flagAlwaysReload );
      }
    } else {
      if( isWordElement( ch ) )
	word[ wptr++ ] = ch;
      else {
	word[ wptr ] = NULL;
	flagWordHead = TRUE;
	/* lookup */
	contentList = dbFetch( word );
	pos += strlen( word );
	if( NULL == contentList ){
	  printf( "%s", word );
	} else if( NULL == contentList->next ){
	  if( FALSE == flagAlwaysReload
	     && fileLen == strlen( contentList->content->dir )
	     && !strcmp( file, contentList->content->dir ) ){
	    if( count == contentList->content->line )
	      printf( "%s", word );
	    else
	      printf( "<A HREF=#%d>%s</A>",
		     contentList->content->line,
		     word );
	  } else {
	    printf( "<A HREF=%s?-k%s+-l%d+%s>%s</A>",
		   CGI_NAME,
		   word,
		   contentList->content->line,
		   contentList->content->dir,
		   word );
	  }
	} else {
	  printf( "<A HREF=%s?-s+-k%s>%s</A>",
		 CGI_NAME,
		 word,
		 word );
	}
	/* forgotten ``ch'' */
	htmlChar( ch, pos, line, count, flagLineNumber, flagAlwaysReload );
      }
    }
  }
}

main( argc, argv )
  int argc;
  char **argv;
{
  char *key = NULL;
  char *file = NULL;
  char *dbfile = NULL;
  char *program;
  char buf[ BUF_SIZE ];
  int command = DB_HTML;
  int line = 0;
  int flagPartial = FALSE;
  int flagLineNumber = FALSE;
  int flagAlwaysReload = FALSE;
  content_list_t *contentList = NULL;
  query_reply_t *queryReply;

  program = *argv;
  while( *(++argv) ){
    if( '-' == **argv ){
      switch( *(*argv + 1 ) ){
      case 'h':
	printf( "Content-type: text/plain\n\n" );
	printf( "All rights reserved. Copyright (C) 1996 by narita@mt.cs.keio.ac.jp\n\n" );
	printf( "Name: %s\nVersion: %s\n\n", program, VERSION );
	printf( "Usage: %% %s [-c|-r|-t|-o|-D|-d|-k|-l|-s|-n|-m|-f|-p] {FILE or KEY}\n", CGI_NAME );
	printf( "\t-c<cgi name>\n" );
	printf( "\t-r<searcher href>\n" );
	printf( "\t-t<searcher title>\n" );
	printf( "\t-o<source tree origin>\n" );
	printf( "\t-D<default database>\n" );
	printf( "\t-d<database>\n" );
	printf( "\t-k<key>\n" );
	printf( "\t-l<line>\n" );
	printf( "\t-s: indicates entry selection\n" );
	printf( "\t-n: creates new database\n" );
	printf( "\t-m: makes new tags merge to old database\n" );
	printf( "\t-N: adds a line number for each line\n" );
	printf( "\t-R: forces CGI output always reloaded\n" );
	printf( "\t-f: forces simple fetch\n" );
	printf( "\t-p: allows partial matching\n" );
	printf( "\tFILE or KEY: tags file or fetch key\n" );
	exit( 0 );
      case 'k':
	key = *argv+2;
	break;
      case 'D':
	DEFAULT_DATABASE = *argv+2;
	dbfile = DEFAULT_DATABASE;
	break;
      case 'd':
	dbfile = *argv+2;
	break;
      case 'l':
	line = atoi( *argv+2 );
	break;
      case 'n':
	/* new database */
	command = DB_LOAD;
	break;
      case 'm':
	/* merge */
	command = DB_MERGE;
	break;
      case 's':
	/* selection */
	command = DB_SELECT;
	break;
      case 'f':
	/* fetch */
	command = DB_FETCH;
	break;
      case 'p':
	flagPartial = TRUE;
	break;
      case 'c':
	CGI_NAME = *argv+2;
	break;
      case 'r':
	SEARCHER_REF = *argv+2;
	break;
      case 't':
	SEARCHER_TITLE = *argv+2;
	break;
      case 'o':
	DEFAULT_ORIGIN = *argv+2;
	break;
      case 'N':
	flagLineNumber = TRUE;
	break;
      case 'R':
	flagAlwaysReload = TRUE;
	break;
      default:
	printf( "Content-type: text/plain\n\n" );
	printf( "Unknown option: ``%s''\n", *argv );
	exit( 0 );
      }
    } else {
      file = *argv;
    }
  }

  if( DB_SELECT != command && NULL == file && getenv( "QUERY_STRING" ) ){

    /*
     * assume that this cgi is issued by FROM GET METHOD
     */

    queryReply = readQuery();

    dbfile = queryReply->dbfile;
    key    = queryReply->key;

    if( NULL == queryReply->contentList ){
      if( NULL == queryReply->file[ 0 ] ){
	htmlError( NO_ENTRY, key );
	exit( 0 );
      } else
	file = queryReply->file;
    } else if( NULL == queryReply->contentList->next ){
      command = DB_HTML;
      line    = queryReply->contentList->content->line;
      file    = queryReply->contentList->content->dir;
    } else {
      command = DB_SELECT;
      contentList = queryReply->contentList;
    }
  }

  /*
   * command: DB_LOAD, DB_MERGE
   *
   * Required: file, dbfile
   */
  if( DB_LOAD == command || DB_MERGE == command ){
    if( file == NULL ){
      printf( "No input file.\n" );
      exit( 0 );
    }
    if( NULL == dbfile ){
      printf( "No database.\n" );
      exit( 0 );
    }

    if( DB_LOAD == command )
      touchDatabase( dbfile );

    if( 0 > dbminit( dbfile ) ){
      printf( "Would you make (%s.dir, %s.pag) now? [y/n]: ",
	     dbfile, dbfile );
      if( 'y' != getchar() )
	exit( -1 );
      touchDatabase( dbfile );
      dbminit( dbfile );
    }

    dbMerge( file );

    exit( 0 );
  }

  /*
   * command: DB_FETCH
   *
   * Required: file, (line), dbfile, (key).
   */
  if( DB_FETCH == command ){
    if( NULL == dbfile ){
      printf( "NoDatabase\n" );
      exit( 0 );
    }
      
    if( FALSE == existDatabase( dbfile ) ){
      printf( "NoSuchDatabase %s\n", dbfile );
      exit( 0 );
    }

    if( 0 > dbminit( dbfile ) ){
      printf( "NoSuchDatabase %s\n", dbfile );
      exit( 0 );
    }

    if( NULL == file && NULL == key ){
      printf( "NoKeyword\n" );
      exit( 0 );
    }

    if( NULL == key )
      key = file;

    if( flagPartial )
      contentList = dbSearch( key );
    else
      contentList = dbFetch( key );

    if( NULL == contentList ){
      printf( "NoEntry %s", key );
      exit( 0 );
    }

    contentListPrint( contentList, key );

    exit( 0 );
  }

  /*
   * The following part constructs HTML page.
   */

  if( FALSE == existDatabase( dbfile ) ){
    htmlError( NO_DATABASE, dbfile );
    exit( 0 );
  }

  if( 0 > dbminit( dbfile ) ){
    htmlError( NO_DATABASE, dbfile );
    exit( 0 );
  }

  /*
   * command: DB_SELECT
   *
   * Required: dbfile, key.
   */
  if( DB_SELECT == command ){
    if( NULL == file && NULL == key ){
      htmlError( NO_KEYWORD, NULL );
      exit( 0 );
    }

    if( NULL == key )
      key = file;

    if( NULL == contentList ){
      if( flagPartial )
	contentList = dbSearch( key );
      else
	contentList = dbFetch( key );
      if( NULL == contentList ){
	htmlError( NO_ENTRY, key );
	exit( 0 );
      }
    }

    if( NULL == contentList->next ){
      command = DB_HTML;
      line    = contentList->content->line;
      file    = contentList->content->dir;
    } else {
      printf( "Content-type: text/html\n\n" );
      printf( "<HTML><HEAD><TITLE>Entry Selection</TITLE></HEAD>\n" );
      printf( "<BODY><HR>\n" );
      printf( "<H2>Keyword ``%s'' has %d entries</H2>\n",
	     key,
	     contentListCount( contentList ) );
      printf( "<UL>\n" );
      contentListHTML( contentList, key, dbfile );
      printf( "</UL><HR>\n" );
      printf( "</BODY></HTML>\n" );
      
      exit( 0 );
    }
  }

  /*
   * command: DB_HTML
   *
   * Required: file, (line), dbfile, (key).
   */

  if( DB_HTML == command ){
    if( NULL == file ){
      htmlError( NO_FILE, "" );
      exit( 0 );
    }

    strcpy( buf, DEFAULT_ORIGIN );
    strcat( buf, file );

    if( 0 > access( buf, R_OK ) || FALSE == legalFile( file ) ){
      htmlError( NO_FILE, buf );
      exit( 0 );
    }

    printf( "Content-type: text/html\n\n" );
    printf( "<HTML><HEAD><TITLE>%s</TITLE></HEAD>\n", file );
    printf( "<BODY><HR>\n" );
    printf( "<H2>``%s''</H2>\n", file );

    if( key && key[ 0 ] && line > 1 )
      printf( "<H3><UL><LI><A HREF=#%d>Jump to: ``%s'' (line:%d)</A></UL></H3>\n", line, key, line );

    printf( "<H3><UL><LI><A HREF=%s>Return to: ``%s''</A></UL></H3>\n", SEARCHER_REF, SEARCHER_TITLE );

    printf( "<HR>\n<PRE>" );
    load( file, buf, dbfile, line, flagLineNumber, flagAlwaysReload );
    printf( "</PRE><HR></BODY></HTML>\n" );
  }
}
